import * as v from 'valibot';
import {getObjectSchema} from './utils';
import {
	InvalidOperationError,
	InvalidPathValueError,
	InvalidSchemaInPathError,
	PathValueTestFailedError,
} from './error';
import {append, get, remove, set} from './object';

export const DELTA_SCHEMA = v.union([
	v.object({
		op: v.literal('add'),
		path: v.string(), // todo validate if valid path?
		value: v.unknown() // todo validate if valid value?
	}),
	v.object({
		op: v.literal('remove'),
		path: v.string(),
	}),
	v.object({
		op: v.literal('replace'),
		path: v.string(), // todo validate if valid path?
		value: v.unknown() // todo validate if valid value?
	}),
	v.object({
		op: v.literal('move'),
		path: v.string(),
		from: v.string(),
	}),
	v.object({
		op: v.literal('copy'),
		path: v.string(),
		from: v.string(),
	}),
	v.object({
		op: v.literal('test'),
		path: v.string(), // todo validate if valid path?
		value: v.unknown() // todo validate if valid value?
	}),
]);

export type Delta = v.Output<typeof DELTA_SCHEMA>;

export const applyDelta = async <T extends v.BaseSchema = v.BaseSchema>(
	resourceSchema: T,
	existing: Record<string, unknown>,
	deltaCollection: Delta[],
) => {
	return await deltaCollection.reduce(
		async (resultObject, deltaItem) => {
			const mutablePreviousObject = await resultObject;

			if (resourceSchema.type !== 'object') {
				return mutablePreviousObject;
			}

			const resourceObjectSchema = resourceSchema as unknown as v.ObjectSchema<any>;
			const pathSchema = getObjectSchema(resourceObjectSchema, deltaItem.path);
			if (typeof pathSchema === 'undefined') {
				throw new InvalidSchemaInPathError();
			}

			switch (deltaItem.op) {
				case 'replace': {
					if (!v.is(pathSchema, deltaItem.value)) {
						throw new InvalidPathValueError();
					}

					set(mutablePreviousObject, deltaItem.path, deltaItem.value);
					return mutablePreviousObject;
				}
				case 'add': {
					if (pathSchema.type !== 'array') {
						throw new InvalidOperationError();
					}

					const arraySchema = pathSchema as v.ArraySchema<any>;
					if (!v.is(arraySchema.item, deltaItem.value)) {
						throw new InvalidPathValueError();
					}

					return append(mutablePreviousObject, deltaItem.path, deltaItem.value);
				}
				case 'remove': {
					return remove(mutablePreviousObject, deltaItem.path);
				}
				case 'copy': {
					const value = get(mutablePreviousObject, deltaItem.from);
					if (!v.is(pathSchema, value)) {
						throw new InvalidPathValueError();
					}

					return set(mutablePreviousObject, deltaItem.path, value);
				}
				case 'move': {
					const value = get(mutablePreviousObject, deltaItem.from);
					if (!v.is(pathSchema, value)) {
						throw new InvalidPathValueError();
					}

					remove(mutablePreviousObject, deltaItem.from)
					return set(mutablePreviousObject, deltaItem.path, value);
				}
				case 'test': {
					const value = get(mutablePreviousObject, deltaItem.path);
					if (value !== deltaItem.value) {
						throw new PathValueTestFailedError();
					}

					return mutablePreviousObject;
				}
				default:
					break;
			}

			if (!v.is(resourceObjectSchema, mutablePreviousObject)) {
				throw new InvalidOperationError();
			}

			return mutablePreviousObject;
		},
		Promise.resolve(existing),
	);
};
